#Copyright 2010-2011 Miquel Torres <tobami@googlemail.com>
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#
"""library for parsing and printing role, cookbook and node information"""
import os
import simplejson as json

from fabric import colors
from fabric.api import env, settings
from fabric.utils import abort

from littlechef.settings import cookbook_paths


def get_nodes():
    """Gets all nodes found in the nodes/ directory"""
    if not os.path.exists('nodes/'):
        return []
    nodes = []
    for filename in sorted([f for f in os.listdir('nodes/')
                                if not os.path.isdir(f) and ".json" in f
                                    and not f.startswith('.')]):
        hostname = ".".join(filename.split('.')[:-1])#remove .json from name
        node = get_node(hostname)
        # Add node name so that we can tell to which node the data belongs to
        node['name'] = hostname
        nodes.append(node)
    return nodes


def get_node(name):
    node_path = os.path.join("nodes", name + ".json")
    if not os.path.exists(node_path):
        return {'run_list': []}
    # Read node.json
    with open(node_path, 'r') as f:
        try:
            node = json.loads(f.read())
        except json.decoder.JSONDecodeError as e:
            msg = 'LittleChef found the following error in'
            msg += ' "{0}":\n                {1}'.format(node_path, str(e))
            abort(msg)
    return node


def print_node(node, detailed=False):
    """Pretty prints the given node"""
    nodename = node['name']
    print(colors.yellow("\n" + nodename))
    # Roles
    if detailed:
        for role in get_roles_in_node(node):
            print_role(_get_role(role), detailed=False)
    else:
        print('  Roles: {0}'.format(", ".join(get_roles_in_node(node))))
    # Recipes
    if detailed:
        for recipe in get_recipes_in_node(node):
            print "  Recipe:", recipe
            print "    attributes: {0}".format(node.get(recipe, ""))
    else:
        print('  Recipes: {0}'.format(", ".join(get_recipes_in_node(node))))
    # Node attributes
    print "  Node attributes:"
    for attribute in node.keys():
        if attribute == "run_list" or attribute == "littlechef":
            continue
        print "    {0}: {1}".format(attribute, node[attribute])


def get_recipes_in_cookbook(name):
    """Gets the name of all recipes present in a cookbook"""
    recipes = []
    path = None
    cookbook_exists = False
    metadata_exists = False
    for cookbook_path in cookbook_paths:
        path = os.path.join(cookbook_path, name)
        # cookbook exists if present in any of the cookbook paths
        cookbook_exists = cookbook_exists or os.path.exists(path)
        # Now try to open metadata.json
        try:
            with open(os.path.join(path, 'metadata.json'), 'r') as f:
                try:
                    cookbook = json.loads(f.read())
                except json.decoder.JSONDecodeError as e:
                    msg = "Little Chef found the following error in your"
                    msg += " {0}.json file:\n  {1}".format(
                        os.path.join(path, 'metadata.json'), e)
                    abort(msg)
                # Add each recipe defined in the cookbook
                metadata_exists = True
                for recipe in cookbook.get('recipes', []):
                    recipes.append({
                        'name': recipe,
                        'description': cookbook['recipes'][recipe],
                        'version': cookbook.get('version'),
                        'dependencies': cookbook.get('dependencies', {}).keys(),
                        'attributes': cookbook.get('attributes', {}).keys(),
                        })
                # When a recipe has no default recipe (libraries?),
                # add one so that it is listed
                if not recipes:
                    recipes.append({
                        'name': name,
                        'description': 'This cookbook has no default recipe',
                        'version': cookbook.get('version'),
                        'dependencies': cookbook.get('dependencies', {}).keys(),
                        'attributes': cookbook.get('attributes', {}).keys(),})
            # Cookbook metadata.json was found, don't try next cookbook path
            # because metadata.json in site-cookbooks has preference
            break
        except IOError:
            # metadata.json was not found, try next cookbook_path
            pass
    if not cookbook_exists:
        abort('Unable to find cookbook "{0}"'.format(name))
    elif not metadata_exists:
        abort('Cookbook "{0}" has no metadata.json'.format(name))
    return recipes


def get_recipes_in_node(node):
    """Gets the name of all recipes present in the run_list of a node"""
    recipes = []
    for elem in node.get('run_list'):
        if elem.startswith("recipe"):
            recipe = elem.split('[')[1].split(']')[0]
            recipes.append(recipe)
    return recipes


def get_recipes():
    """Gets all recipes found in the cookbooks/ directory"""
    recipes = []
    for dirname in sorted(
        [d for d in os.listdir('cookbooks') if os.path.isdir(
            os.path.join('cookbooks', d)) and not d.startswith('.')]):
        recipes.extend(get_recipes_in_cookbook(dirname))
    return recipes


def print_recipe(recipe):
    """Pretty prints the given recipe"""
    print(colors.yellow("\n{0}".format(recipe['name'])))
    print "  description:  {0}".format(recipe['description'])
    print "  version:      {0}".format(recipe['version'])
    print "  dependencies: {0}".format(", ".join(recipe['dependencies']))
    print "  attributes:   {0}".format(", ".join(recipe['attributes']))


def get_roles_in_node(node):
    """Gets the name of all roles found in the run_list of a node"""
    roles = []
    for elem in node.get('run_list'):
        if elem.startswith("role"):
            role = elem.split('[')[1].split(']')[0]
            roles.append(role)
    return roles


def _get_role(rolename):
    """Reads and parses a file containing a role"""
    path = 'roles/' + rolename + '.json'
    if not os.path.exists(path):
        abort("Couldn't read role file {0}".format(path))
    with open(path, 'r') as f:
        try:
            role = json.loads(f.read())
        except json.decoder.JSONDecodeError as e:
            msg = "Little Chef found the following error in your"
            msg += " {0}.json file:\n  {1}".format(rolename, str(e))
            abort(msg)
        role['fullname'] = rolename
        return role


def get_roles():
    """Gets all roles found in the roles/ directory"""
    roles = []
    for root, subfolders, files in os.walk('roles/'):
        for filename in files:
            if filename.endswith(".json"):
                path = os.path.join(
                    root[len('roles/'):], filename[:-len('.json')])
                roles.append(_get_role(path))
    return roles


def print_role(role, detailed=True):
    """Pretty prints the given role"""
    if detailed:
        print(colors.yellow(role.get('fullname')))
    else:
        print("  Role: {0}".format(role.get('fullname')))
    if detailed:
        print("    description: {0}".format(role.get('description')))
    if 'default_attributes' in role:
        print("    default_attributes:")
        _pprint(role['default_attributes'])
    if 'override_attributes' in role:
        print("    override_attributes:")
        _pprint(role['override_attributes'])
    print("")


def get_cookbook_path(cookbook_name):
    """Returns path to the cookbook for the given cookbook name"""
    for cookbook_path in cookbook_paths:
        path = os.path.join(cookbook_path, cookbook_name)
        if os.path.exists(path):
            return path
    raise IOError('Can\'t find cookbook with name "{0}"'.format(cookbook_name))


def _pprint(dic):
    """Prints a dictionary with one indentation level"""
    for key, value in dic.items():
        print("        {0}: {1}".format(key, value))


def print_header(string):
    """Prints a colored header"""
    print(colors.yellow("\n== {0} ==".format(string)))


def get_margin(length):
    """Add enough tabs to align in two columns"""
    if length > 23:
        margin_left = "\t"
        chars = 1
    elif length > 15:
        margin_left = "\t\t"
        chars = 2
    elif length > 7:
        margin_left = "\t\t\t"
        chars = 3
    else:
        margin_left = "\t\t\t\t"
        chars = 4
    return margin_left


def credentials(*args, **kwargs):
    """Override default credentials with contents of .ssh/config,
    if appropriate
    """
    if env.ssh_config:
        credentials = env.ssh_config.lookup(env.host)
        # translate from paramiko params to fabric params
        if 'identityfile' in credentials:
            credentials['key_filename'] = credentials['identityfile']
        credentials.update(kwargs)
    else:
        credentials = kwargs
    return settings(*args, **credentials)



#Copyright 2010-2011 Miquel Torres <tobami@googlemail.com>
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#
import unittest
import os
import shutil
import json
from os.path import join, split, sep, normpath, abspath, exists

from fabric.api import *

import chef
import lib
import runner


# Set paths
littlechef_src = split(normpath(abspath(__file__)))[0]
littlechef_top = normpath(join(littlechef_src, '..'))
littlechef_tests = join(littlechef_top, 'tests')


class BaseTest(unittest.TestCase):
    def setUp(self):
        """Simulate we are inside a kitchen"""
        # Orient ourselves
        os.chdir(littlechef_src)
        for d in ['nodes', 'roles', 'cookbooks', 'data_bags']:
            shutil.copytree(join(littlechef_tests, '{0}'.format(d)), d)
        shutil.copy(join(littlechef_tests, 'auth.cfg'), littlechef_src)

    def tearDown(self):
        os.chdir(littlechef_src)
        for d in ['nodes', 'roles', 'cookbooks', 'data_bags']:
            shutil.rmtree(d)
        if exists('tmp_node.json'):
            os.remove('tmp_node.json')
        os.remove('auth.cfg')


class TestRunner(BaseTest):
    def test_not_a_kitchen(self):
        """Should exit with error when not a kitchen directory"""
        # Change to a directory which is not a kitchen
        os.chdir(littlechef_top)
        self.assertRaises(SystemExit, runner._readconfig)

    def test_readconfig(self):
        """Should read auth.cfg and properly configure variables"""
        runner._readconfig()
        self.assertEquals(env.password, 'testpass')


class TestLib(BaseTest):
    def test_get_node(self):
        """Should get data for a given node, empty when it doesn't exist"""
        expected = {'run_list': []}
        self.assertEquals(lib.get_node('Idon"texist'), expected)
        expected = {'run_list': ['recipe[subversion]']}
        self.assertEquals(lib.get_node('testnode'), expected)

    def test_list_nodes(self):
        """Should list all configured nodes"""
        expected = [{'name': 'testnode', 'run_list': ['recipe[subversion]']}]
        self.assertEquals(lib.get_nodes(), expected)

    def test_list_recipes(self):
        recipes = lib.get_recipes()
        self.assertEquals(len(recipes), 3)
        self.assertEquals(recipes[1]['description'],
            'Subversion Client installs subversion and some extra svn libs')
        self.assertEquals(recipes[2]['name'], 'subversion::server')


class TestChef(BaseTest):
    def test_save_config(self):
        """Should create tmp_node.json and a nodes/testnode2.json config file"""
        env.host_string = 'testnode2'
        run_list = ["role[testrole]"]
        chef._save_config({"run_list": run_list})
        self.assertTrue(exists(join('nodes', 'testnode2.json')))
        with open(join('nodes','testnode2.json'), 'r') as f:
            data = json.loads(f.read())
            self.assertEquals(data['run_list'], run_list)
        # It should't overwrite existing config files
        env.host_string = 'testnode'# This node exists
        run_list = ["role[testrole]"]
        chef._save_config({"run_list": run_list})
        with open(join('nodes','testnode.json'), 'r') as f:
            data = json.loads(f.read())
            # It should *NOT* have "testrole" assigned
            self.assertEquals(data['run_list'], ["recipe[subversion]"])

    def test_build_node(self):
        """Should build cookbooks dependencies"""
        env.host_string = 'testnode'
        cookbooks = chef._build_node(lib.get_node(env.host_string))
        self.assertEquals(cookbooks, ['subversion'])
        #TODO: add more cookbooks with dependencies, add apache2


if __name__ == "__main__":
    unittest.main()



#Copyright 2010-2011 Miquel Torres <tobami@googlemail.com>
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#
# Path to look up cookbooks
cookbook_paths = ['site-cookbooks', 'cookbooks']

# Node's Chef Solo working directory for storing cookbooks, roles, etc.
node_work_path = '/var/chef-solo'



"""LittleChef's setup.py"""
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import platform


scripts = ['cook']
if platform.system() == 'Windows':
    scripts += ['cook.cmd']

setup(
    name="littlechef",
    version=__import__('littlechef').version,
    description="Cook with Chef without a Chef Server",
    author="Miquel Torres",
    author_email="tobami@googlemail.com",
    url="http://github.com/tobami/littlechef",
    download_url="http://github.com/tobami/littlechef/archives/master",
    keywords=["chef", "devops"],
    install_requires=['fabric>=1.0.1', 'simplejson'],
    packages=['littlechef'],
    package_data={'littlechef': ['data_bags_patch.rb']},
    scripts=['cook'],
    test_suite='nose.collector',
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Topic :: System :: Systems Administration',
        ],
    long_description="""\
Cook with Chef without Chef Server
-------------------------------------
With LittleChef you will get all you need to start cooking with Chef_.

It works as follows: Whenever you apply a recipe to a node, all needed
cookbooks and its dependencies are gzipped and uploaded to that node. A
node.json file gets created on the fly and uploaded, and Chef Solo gets
executed at the remote node, using node.json as the node configuration and the
pre-installed solo.rb for Chef Solo configuration. Cookbooks, data bags and roles
are configured to be found at (/var/chef-solo/).

The result is that you can play as often with your recipes and nodes as you
want, without having to worry about a central Chef repository, Chef server nor
anything else. You can make small changes to your cookbooks and test them again
and again without having to commit the changes. LittleChef brings sanity
to cookbook development.

.. _Chef: http://wiki.opscode.com/display/chef/Home
"""
)
